#include <GameINILoader.h>

#include <FileClasses/FileManager.h>

#include <GameClass.h>
#include <AiPlayerClass.h>
#include <PlayerClass.h>
#include <MapClass.h>
#include <MapGenerator.h>
#include <ScreenBorder.h>
#include <MapSeed.h>
#include <GameInitInfoClass.h>

#include <structures/StructureClass.h>
#include <units/UnitClass.h>
#include <structures/BuilderClass.h>

#include <Trigger/ReinforcementTrigger.h>
#include <Trigger/TimeoutTrigger.h>

#include <misc/string_util.h>

#include <sand.h>
#include <globals.h>

#include <stdio.h>
#include <stdexcept>
#include <algorithm>

GameINILoader::GameINILoader(GameClass* pGameClass, std::string mapname)
{
    this->pGameClass = pGameClass;
    this->mapname = mapname;

    SDL_RWops* mapiniFile = NULL;
    try {
        mapiniFile = pFileManager->OpenFile(this->mapname);
        inifile = std::shared_ptr<INIFile>(new INIFile(mapiniFile));
        SDL_RWclose(mapiniFile);
    } catch (...) {
        if(mapiniFile != NULL) {
            SDL_RWclose(mapiniFile);
        }
        throw;
    }

    load();
}

GameINILoader::~GameINILoader() {
}


/**
	Loads a map from an INI-File.
*/
void GameINILoader::load()
{
    loadMap();
    loadPlayers();
    loadUnits();
    loadStructures();
    loadReinforcements();
    loadView();
    loadChoam();
}

/**
    This method loads the game map. This is based on the [MAP] section in the INI file.
*/
void GameINILoader::loadMap() {
	currentGameMap = new MapClass(64, 64);

	currentGame->winFlags = inifile->getIntValue("BASIC","WinFlags",3);
	currentGame->loseFlags = inifile->getIntValue("BASIC","LoseFlags",1);

	int timeout = inifile->getIntValue("BASIC","TIMEOUT",0);

	if((timeout != 0) && ((currentGame->winFlags & WINLOSEFLAGS_TIMEOUT) != 0)) {
	    std::shared_ptr<Trigger> newTrigger = std::shared_ptr<Trigger>(new TimeoutTrigger(MILLI2CYCLES(timeout * 60 * 1000)));
        currentGame->GetTriggerManager().addTrigger(newTrigger);
	}

	int SeedNum = inifile->getIntValue("MAP","Seed",-1);

	if(SeedNum == -1) {
		delete currentGameMap;
		throw std::runtime_error("GameINILoader::loadMap(): Cannot find Seednum in" + mapname + "!");
	}

	unsigned short SeedMap[64*64];
	createMapWithSeed(SeedNum,SeedMap);

	for (int j = 0; j < currentGameMap->sizeY; j++) {
		for (int i = 0; i < currentGameMap->sizeX; i++) {
			int type = Terrain_Sand;
			unsigned char seedmaptype = SeedMap[j*64+i] >> 4;
			switch(seedmaptype) {

				case 0x7:   /* Sand */
					type = Terrain_Sand;
					break;

				case 0x2:   /* Building */
				case 0x8:	/* Rock */
					type = Terrain_Rock;
					break;

				case 0x9:   /* Dunes */
					type = Terrain_Dunes;
					break;

                case 0xa:   /* Mountain */
					type = Terrain_Mountain;
					break;

				case 0xb:	/* Spice */
					type = Terrain_Spice;
					break;

				case 0xc:   /* ThickSpice */
					type = Terrain_ThickSpice;
					break;

				default:
                    throw std::runtime_error("GameINILoader::loadMap(): Unknown maptype!");
                    break;
			}

			currentGameMap->cell[i][j].setType(type);
		}

	}

	currentGameMap->createSandRegions();

	std::string BloomString = inifile->getStringValue("MAP","Bloom");
	if(BloomString != "") {
		std::vector<std::string> BloomPositions  = SplitString(BloomString);

		for(unsigned int i=0; i < BloomPositions.size();i++) {
			// set bloom
			int BloomPos = atol(BloomPositions[i].c_str());
			if((BloomPos != 0) || (BloomPositions[i] == "0")) {
				int xpos = BloomPos % currentGameMap->sizeX;
				int ypos = BloomPos / currentGameMap->sizeX;
				if(currentGameMap->cellExists(xpos,ypos) && currentGameMap->getCell(xpos,ypos)->getType() == Terrain_Sand) {
					currentGameMap->cell[xpos][ypos].setType(Terrain_SpiceBloom);
				} else {
					fprintf(stderr,"GameINILoader::loadMap(): Cannot set bloom at %d, %d\n",xpos, ypos);
				}
			}
		}

	}

	std::string SpecialString = inifile->getStringValue("MAP","Special");
	if(SpecialString != "") {
		std::vector<std::string> SpecialPositions  = SplitString(SpecialString);

		for(unsigned int i=0; i < SpecialPositions.size();i++) {
			// set special
			int SpecialPos = atol(SpecialPositions[i].c_str());
			if((SpecialPos != 0) || (SpecialPositions[i] == "0")) {
				int xpos = SpecialPos % currentGameMap->sizeX;
				int ypos = SpecialPos / currentGameMap->sizeX;
				if(currentGameMap->cellExists(xpos,ypos) && currentGameMap->getCell(xpos,ypos)->getType() == Terrain_Sand) {
					currentGameMap->cell[xpos][ypos].setType(Terrain_SpecialBloom);
				} else {
					fprintf(stderr,"GameINILoader::loadMap(): Cannot set special at %d, %d\n",xpos, ypos);
				}
			}
		}

	}

	std::string FieldString = inifile->getStringValue("MAP","Field");
	if(FieldString != "") {
		std::vector<std::string> FieldPositions  = SplitString(FieldString);

		for(unsigned int i=0; i < FieldPositions.size();i++) {
			// set bloom
			int FieldPos = atol(FieldPositions[i].c_str());
			if((FieldPos != 0) || (FieldPositions[i] == "0")) {
				int xpos = FieldPos % currentGameMap->sizeX;
				int ypos = FieldPos / currentGameMap->sizeX;
				if(currentGameMap->cellExists(xpos,ypos)) {
                    for(int i = -5; i <= 5; i++) {
                        for(int j = -5; j <= 5; j++) {
                            if(currentGameMap->cellExists(xpos + i, ypos + j)) {
                                TerrainClass* cell = currentGameMap->getCell(xpos + i, ypos + j);

                                if(cell->isSand() && (distance_from(xpos, ypos, xpos + i, ypos + j) <= 5)) {
                                    cell->setType(((i==0)&&(j==0)) ? Terrain_ThickSpice : Terrain_Spice);
                                }
                            }
                        }
                    }
				} else {
					fprintf(stderr,"GameINILoader::loadMap(): Cannot set field at %d, %d\n",xpos, ypos);
				}
			}
		}

	}
}

/**
    This method loads the players on the map specified by the various player sections in the INI file ([Atreides], [Ordos], [Harkonnen]).
*/
void GameINILoader::loadPlayers()
{
	// now set up all the players
	for(int house=0; house<NUM_HOUSES;house++) {
        std::string houseName = getHouseNameByNumber((HOUSETYPE) house);

        if(inifile->hasSection(houseName)) {
            int startingCredits = inifile->getIntValue(houseName,"Credits",DEFAULT_STARTINGCREDITS);
            int quota = inifile->getIntValue(houseName,"Quota",0);
            if(pGameClass->InitSettings->House != house) {
                pGameClass->player[house] = new AiPlayerClass(house,house,house,startingCredits,pGameClass->InitSettings->Difficulty, 2, quota);
            } else {
                pGameClass->player[house] = new PlayerClass(house,house,house,startingCredits, 1, quota);
                thisPlayer = pGameClass->player[house];
            }
        }
	}
}

/**
    This method loads the choam section of the INI file
*/
void GameINILoader::loadChoam()
{
    INIFile::KeyListHandle myListHandle = inifile->KeyList_Open("CHOAM");

    while(!inifile->KeyList_EOF(myListHandle)) {
        std::string UnitStr = inifile->KeyList_GetNextKey(&myListHandle);

        Uint32 unitID = getItemIDByName(UnitStr);
        if((unitID == ItemID_Invalid) || !isUnit(unitID)) {
            throw std::runtime_error("GameINILoader::loadChoam(): Invalid unit string: " + UnitStr);
        }

        int num = inifile->getIntValue("CHOAM",UnitStr,-2);
        if(num == -2) {
            throw std::runtime_error("GameINILoader::loadChoam(): Invalid number");
        }

        if(num == -1) {
            num = 0;
        }

        for(int i=0;i<MAX_PLAYERS;i++) {
            if(currentGame->player[i] != NULL) {
                currentGame->player[i]->getChoam().addItem(unitID, num);
            }
        }
    }
}

/**
    This method loads the units specified by the [Units] section.
*/
void GameINILoader::loadUnits()
{
	INIFile::KeyListHandle myListHandle = inifile->KeyList_Open("UNITS");

	while(!inifile->KeyList_EOF(myListHandle)) {
		std::string tmpkey = inifile->KeyList_GetNextKey(&myListHandle);
		std::string tmp = inifile->getStringValue("UNITS",tmpkey);
		std::string HouseStr, UnitStr, health, PosStr, rotation, mode;
		SplitString(tmp,6,&HouseStr,&UnitStr,&health,&PosStr,&rotation,&mode);

        int house = getHouseByName(HouseStr);
        if(house == HOUSE_INVALID) {
            throw std::runtime_error("GameINILoader::loadUnits(): Invalid house string: " + HouseStr);
        }

		int pos = atol(PosStr.c_str());

		if(pos <= 0) {
		    throw std::runtime_error("GameINILoader::loadUnits(): Invalid position string: " + PosStr);
		}

		int Num2Place = 1;
		int unitID;

		if(UnitStr == "Infantry")	{
			//make three
			unitID = Unit_Soldier;
			Num2Place = 3;
		} else if (UnitStr == "Troopers") {
			//make three
			unitID = Unit_Trooper;
			Num2Place = 3;
		} else {
            unitID = getItemIDByName(UnitStr);
            if((unitID == ItemID_Invalid) || !isUnit(unitID)) {
                throw std::runtime_error("GameINILoader::loadUnits(): Invalid unit string: " + UnitStr);
            }
		}

		double percentHealth = std::min(((double) atol(health.c_str())) / 256.0, 1.0);

		ATTACKMODE attackmode = getAttackModeByName(mode);
		if(attackmode == ATTACKMODE_INVALID) {
            throw std::runtime_error("GameINILoader::loadUnits(): Invalid attackmode string: " + mode);
		}

		for(int i = 0; i < Num2Place; i++) {
			UnitClass* newUnit = getOrCreatePlayer(house)->placeUnit(unitID, pos%64, pos/64);
			if(newUnit == NULL) {
				fprintf(stderr, "GameINILoader::loadUnits(): Invalid or occupied position: %d!\n", pos);
			} else {
			    newUnit->setHealth(lround(newUnit->getMaxHealth() * percentHealth));
                newUnit->DoSetAttackMode(attackmode);
			}
		}
	}

	inifile->KeyList_Close(&myListHandle);
}

/**
    This method loads the structures specified by the [Structures] section.
*/
void GameINILoader::loadStructures()
{
	INIFile::KeyListHandle myListHandle = inifile->KeyList_Open("STRUCTURES");

	while(!inifile->KeyList_EOF(myListHandle)) {
		std::string tmpkey = inifile->KeyList_GetNextKey(&myListHandle);
		std::string tmp = inifile->getStringValue("STRUCTURES",tmpkey);

		if(tmpkey.find("GEN") == 0) {
			// Gen Object/Structure
			std::string PosStr = tmpkey.substr(3,tmpkey.size()-3);
			int pos = atol(PosStr.c_str());

			std::string HouseStr, BuildingStr;
			SplitString(tmp,2,&HouseStr,&BuildingStr);

			int house = getHouseByName(HouseStr);
			if(house == HOUSE_INVALID) {
				throw std::runtime_error("GameINILoader::loadStructures(): Invalid house string: " + HouseStr);
			}

			if(BuildingStr == "Concrete") {
				getOrCreatePlayer(house)->placeStructure(NONE, Structure_Slab1, pos%64, pos/64);
			} else if(BuildingStr == "Wall") {
				getOrCreatePlayer(house)->placeStructure(NONE, Structure_Wall, pos%64, pos/64);
			} else {
			    throw std::runtime_error("GameINILoader::loadStructures(): Invalid building string: " + BuildingStr);
			}
		} else {
			// other structure
			std::string HouseStr, BuildingStr, health, PosStr;
			SplitString(tmp,6,&HouseStr,&BuildingStr,&health,&PosStr);

			int pos = atol(PosStr.c_str());

			int house = getHouseByName(HouseStr);
			if(house == HOUSE_INVALID) {
				throw std::runtime_error("GameINILoader::loadStructures(): Invalid house string: " + HouseStr);
			}

			double percentHealth = std::min(((double) atol(health.c_str())) / 256.0, 1.0);

			int itemID = getItemIDByName(BuildingStr);

			if((itemID == ItemID_Invalid) || !isStructure(itemID)) {
				throw std::runtime_error("GameINILoader::loadStructures(): Invalid building string: " + BuildingStr);
			}

			if (itemID != 0) {
				ObjectClass* newStructure = getOrCreatePlayer(house)->placeStructure(NONE, itemID, pos%64, pos/64);
				if(newStructure == NULL) {
					fprintf(stderr,"GameINILoader::loadStructures(): Invalid or occupied position: %s\n",PosStr.c_str());
				} else {
				    newStructure->setHealth(lround(newStructure->getMaxHealth() * percentHealth));
				}
			}
		}
	}

	inifile->KeyList_Close(&myListHandle);
}

/**
    This method loads the reinforcements from the [REINFORCEMENTS] section.
*/
void GameINILoader::loadReinforcements()
{
	INIFile::KeyListHandle myListHandle = inifile->KeyList_Open("REINFORCEMENTS");

	while(!inifile->KeyList_EOF(myListHandle)) {
		std::string currentKey = inifile->KeyList_GetNextKey(&myListHandle);
		std::string currentValue = inifile->getStringValue("REINFORCEMENTS",currentKey);

		std::string strHouseName;
		std::string strUnitName;
		std::string strDropLocation;
		std::string strTime;
		std::string strPlus;

		std::vector<std::string> valueList = SplitString(currentValue);

        if(SplitString(currentValue, 4, &strHouseName, &strUnitName, &strDropLocation, &strTime) == false) {
            if(SplitString(currentValue, 5, &strHouseName, &strUnitName, &strDropLocation, &strTime, &strPlus) == false) {
                throw std::runtime_error("GameINILoader::loadReinforcements(): Invalid reinforcement string: " + currentKey + " = " + currentValue);
            }
        }

        int house = getHouseByName(strHouseName);
        if(house == HOUSE_INVALID) {
            throw std::runtime_error("GameINILoader::loadReinforcements(): Invalid house string \"" + strHouseName + "\": " + currentKey + " = " + currentValue);
        }

        int Num2Drop = 1;
        Uint32 unitID = 0;
        if(strUnitName == "Infantry")	{
			//make three
			unitID = Unit_Soldier;
			Num2Drop = 3;
		} else if (strUnitName == "Troopers") {
			//make three
			unitID = Unit_Trooper;
			Num2Drop = 3;
		} else {
            unitID = getItemIDByName(strUnitName);
            if((unitID == ItemID_Invalid) || !isUnit(unitID)) {
                throw std::runtime_error("GameINILoader::loadReinforcements(): Invalid unit string \"" + strUnitName + "\": " + currentKey + " = " + currentValue);
            }
		}

        // convert to lower case
        std::transform(strDropLocation.begin(), strDropLocation.end(), strDropLocation.begin(), (int(*)(int)) tolower);

        ReinforcementTrigger::DropLocation dropLocation;
        if(strDropLocation == "north") {
            dropLocation = ReinforcementTrigger::Drop_North;
        } else if(strDropLocation == "east") {
            dropLocation = ReinforcementTrigger::Drop_East;
        } else if(strDropLocation == "south") {
            dropLocation = ReinforcementTrigger::Drop_South;
        } else if(strDropLocation == "west") {
            dropLocation = ReinforcementTrigger::Drop_West;
        } else if(strDropLocation == "air") {
            dropLocation = ReinforcementTrigger::Drop_Air;
        } else if(strDropLocation == "visible") {
            dropLocation = ReinforcementTrigger::Drop_Visible;
        } else if(strDropLocation == "enemybase") {
            dropLocation = ReinforcementTrigger::Drop_Enemybase;
        } else if(strDropLocation == "homebase") {
            dropLocation = ReinforcementTrigger::Drop_Homebase;
        } else {
            throw std::runtime_error("GameINILoader::loadReinforcements(): Invalid drop location string \"" + strDropLocation + "\": " + currentKey + " = " + currentValue);
        }

        Uint32 droptime = atol(strTime.c_str());
        if(droptime == 0 && strTime != "0" && strTime != "0+") {
            throw std::runtime_error("GameINILoader::loadReinforcements(): Invalid drop time string \"" + strTime + "\": " + currentKey + " = " + currentValue);
        }
        Uint32 dropCycle = MILLI2CYCLES(droptime * 60 * 1000);

        bool bRepeat = (strTime.rfind('+') == (strTime.length() - 1)) || (strPlus == "+");

        for(int i=0;i<Num2Drop;i++) {
            // check if there is a similar trigger at the same time
            const std::list< std::shared_ptr<Trigger> >& triggerList = currentGame->GetTriggerManager().getTriggers();
            std::list< std::shared_ptr<Trigger> >::const_iterator iter;

            bool bInserted = false;
            for(iter = triggerList.begin(); iter != triggerList.end(); ++iter) {
                Trigger *pTrigger = iter->get();

                ReinforcementTrigger* pReinforcementTrigger = dynamic_cast<ReinforcementTrigger*>(pTrigger);

                if(pReinforcementTrigger != NULL
                    && pReinforcementTrigger->getCycleNumber() == dropCycle
                    && pReinforcementTrigger->getHouse() == house
                    && pReinforcementTrigger->isRepeat() == bRepeat
                    && pReinforcementTrigger->getDropLocation() == dropLocation) {

                    // add the new reinforcement to this reinforcement (call only one carryall)
                    pReinforcementTrigger->addUnit(unitID);
                    bInserted = true;
                    break;
                }
            }

            if(bInserted == false) {
                std::shared_ptr<Trigger> newTrigger = std::shared_ptr<Trigger>(new ReinforcementTrigger(house, unitID, dropLocation, bRepeat, dropCycle));
                currentGame->GetTriggerManager().addTrigger(newTrigger);
            }
        }
	}

	inifile->KeyList_Close(&myListHandle);
}

/**
    This method sets up the view specified by "TacticalPos" in the [BASIC] section.
*/
void GameINILoader::loadView()
{
	int tacticalPosInt = inifile->getIntValue("BASIC","TacticalPos",-100) + 64*5 + 7;
	Coord tacticalPos(tacticalPosInt % 64, tacticalPosInt / 64);

	if(tacticalPos.x < 0 || tacticalPos.x >= 64 || tacticalPos.y < 0 || tacticalPos.y > 64) {
        fprintf(stderr,"Warning: Invalid tacticalPosInt!\n");
        pGameClass->setupView();
	} else {
        screenborder->SetNewScreenCenter(tacticalPos*BLOCKSIZE);
	}
}

/**
    This method returns the player with house. If it does not already exist a new AI Player is created and returned.
    \param house the house of the player to return or create
    \return the player specified by house
*/
PlayerClass* GameINILoader::getOrCreatePlayer(int house) {
    if(pGameClass->player[house] == NULL) {
        pGameClass->player[house] = new AiPlayerClass(house,house,house,0,pGameClass->InitSettings->Difficulty);
    }

    return pGameClass->player[house];
}
