/*
 *  This file is part of Dune Legacy.
 *
 *  Dune Legacy is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Dune Legacy is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Dune Legacy.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <PlayerClass.h>

#include <globals.h>
#include <sand.h>

#include <FileClasses/TextManager.h>

#include <GameClass.h>
#include <MapClass.h>
#include <RadarView.h>
#include <SoundPlayer.h>
#include <MapGenerator.h>

#include <structures/StructureClass.h>
#include <structures/BuilderClass.h>
#include <structures/RefineryClass.h>
#include <units/Carryall.h>
#include <units/HarvesterClass.h>

#include <algorithm>

PlayerClass::PlayerClass(int newPlayerNumber, int newHouse, int newColour, int newCredits, Uint8 team, int quota) : choam(this)
{
    PlayerClass::init();

	playerNumber = newPlayerNumber;

    house = ((newHouse >= 0) && (newHouse < NUM_HOUSES)) ? newHouse :  0;
    this->team = team;
    colour = ((newColour >= 0) && (newColour < NUM_HOUSES)) ? newColour : 0;

	credits = 0.0;
    startingCredits = newCredits;
    oldCredits = (int) (credits+startingCredits);

    this->quota = quota;

    destroyedValue = 0;
    numDestroyedUnits = 0;
    numDestroyedStructures = 0;
    harvestedSpice = 0.0;
    producedPower = 0;
    powerUsageTimer = 0;
}

PlayerClass::PlayerClass(Stream& stream, int playerNumber) : choam(this)
{
    this->playerNumber = playerNumber;

    PlayerClass::init();

    house = stream.readUint8();
	team = stream.readUint8();
	colour = stream.readUint8();

	credits = stream.readDouble();
	startingCredits = stream.readDouble();
    oldCredits = (int) (credits+startingCredits);
    quota = stream.readSint32();

    destroyedValue = stream.readUint32();
    numDestroyedUnits = stream.readUint32();
    numDestroyedStructures = stream.readUint32();
    harvestedSpice = stream.readDouble();
    producedPower = stream.readSint32();
    powerUsageTimer = stream.readSint32();

    choam.load(stream);

	if(stream.readBool() == true) {
		thisPlayer = this;
	}
}

void PlayerClass::init()
{
    ai = false;

    numUnits = 0;
	numStructures = 0;
	for(int i=0;i<ItemID_max;i++) {
        numItem[i] = 0;
	}

	capacity = 0;
	powerRequirement = 0;
}

PlayerClass::~PlayerClass()
{
}

void PlayerClass::save(Stream& stream) const
{
	stream.writeUint8(house);
	stream.writeUint8(team);
	stream.writeUint8(colour);

	stream.writeDouble(credits);
	stream.writeDouble(startingCredits);
	stream.writeSint32(quota);

    stream.writeUint32(destroyedValue);
    stream.writeUint32(numDestroyedUnits);
    stream.writeUint32(numDestroyedStructures);
    stream.writeDouble(harvestedSpice);
    stream.writeSint32(producedPower);
    stream.writeSint32(powerUsageTimer);

    choam.save(stream);

	stream.writeBool(this == thisPlayer);
}

void PlayerClass::PrintStat() const {
	fprintf(stderr,"Player %d: (Number of Units: %d, Number of Structures: %d)\n",getPlayerNumber(),numUnits,numStructures);
	fprintf(stderr,"Barracks: %d\t\tWORs: %d\n", numItem[Structure_Barracks],numItem[Structure_WOR]);
	fprintf(stderr,"Light Factories: %d\tHeavy Factories: %d\n",numItem[Structure_LightFactory],numItem[Structure_HeavyFactory]);
	fprintf(stderr,"IXs: %d\t\t\tPalaces: %d\n",numItem[Structure_IX],numItem[Structure_Palace]);
	fprintf(stderr,"Repair Yards: %d\t\tHigh-Tech Factories: %d\n",numItem[Structure_RepairYard],numItem[Structure_HighTechFactory]);
	fprintf(stderr,"Refineries: %d\t\tStarports: %d\n",numItem[Structure_Refinery],numItem[Structure_StarPort]);
	fprintf(stderr,"Walls: %d\t\tRocket Turrets: %d\n",numItem[Structure_Wall],numItem[Structure_RocketTurret]);
	fprintf(stderr,"Gun Turrets: %d\t\tConstruction Yards: %d\n",numItem[Structure_GunTurret],numItem[Structure_ConstructionYard]);
	fprintf(stderr,"Windtraps: %d\t\tRadars: %d\n",numItem[Structure_WindTrap],numItem[Structure_Radar]);
	fprintf(stderr,"Silos: %d\n",numItem[Structure_Silo]);
	fprintf(stderr,"Carryalls: %d\t\tFrigates: %d\n",numItem[Unit_Carryall],numItem[Unit_Frigate]);
	fprintf(stderr,"Devastators: %d\t\tDeviators: %d\n",numItem[Unit_Devastator],numItem[Unit_Deviator]);
	fprintf(stderr,"Soldiers: %d\t\tFremen: %d\n",numItem[Unit_Soldier],numItem[Unit_Fremen]);
	fprintf(stderr,"Trooper: %d\t\tDeviators: %d\n",numItem[Unit_Trooper],numItem[Unit_Sardaukar]);
	fprintf(stderr,"Saboteur: %d\t\tSandworms: %d\n",numItem[Unit_Saboteur],numItem[Unit_Sandworm]);
	fprintf(stderr,"Quads: %d\t\tTrikes: %d\n",numItem[Unit_Quad],numItem[Unit_Trike]);
	fprintf(stderr,"Raiders: %d\t\tTanks: %d\n",numItem[Unit_Raider],numItem[Unit_Tank]);
	fprintf(stderr,"Siege Tanks : %d\t\tSonic Tanks: %d\n",numItem[Unit_SiegeTank],numItem[Unit_SonicTank]);
	fprintf(stderr,"Harvesters: %d\t\tMCVs: %d\n",numItem[Unit_Harvester],numItem[Unit_MCV]);
	fprintf(stderr,"Ornithopters: %d\t\tRocket Launchers: %d\n",numItem[Unit_Ornithopter],numItem[Unit_Launcher]);
}

void PlayerClass::addCredits(double newCredits, bool wasRefined)
{
	if (newCredits > 0.0) {
	    if(wasRefined == true) {
            harvestedSpice += newCredits;
	    }

		credits += newCredits;
		if (this == thisPlayer)	{
			if(((currentGame->winFlags & WINLOSEFLAGS_QUOTA) != 0) && (quota != 0))	{
				if (credits >= quota) {
					win();
				}
			}
		}
	}
}

void PlayerClass::checkSelectionLists()
{
    RobustList<StructureClass*>::const_iterator iter;
    for(iter = structureList.begin(); iter != structureList.end(); ++iter) {
		StructureClass* tempStructure = *iter;
        if(tempStructure->isABuilder() && (tempStructure->getOwner() == this)) {
			((BuilderClass*) tempStructure)->checkSelectionList();
        }
    }
}


void PlayerClass::incrementStructures(int itemID)
{
//	fprintf(stderr,"Adding new structure to Player %d: ItemID = %d\n",getPlayerNumber(),itemID);
	numStructures++;
	numItem[itemID]++;

    // change power requirements
	bool bhasPowerBefore = hasPower();
	int currentItemPower = currentGame->objectData.data[itemID].power;
	if(currentItemPower >= 0) {
        powerRequirement += currentGame->objectData.data[itemID].power;
	}
	bool bhasPowerAfterwards = hasPower();

    if(currentGame->gameState != LOADING) {
        if(hasRadar() && (bhasPowerBefore != bhasPowerAfterwards)) {
            changeRadar(bhasPowerAfterwards);
        }
    } else {
        // loading => don't show radar animation
        if(hasRadar()) {
            currentGame->radarView->setRadarMode(bhasPowerAfterwards);
        }
    }

	// change spice capacity
	capacity += currentGame->objectData.data[itemID].capacity;

    if(currentGame->gameState != LOADING) {
        // do not check selection lists if we are loading
        checkSelectionLists();
    }
}


void PlayerClass::decrementStructures(int itemID, const Coord& location)
{
	numStructures--;
    numItem[itemID]--;

	// change power requirements
	bool bhasPowerBefore = hasPower();
	int currentItemPower = currentGame->objectData.data[itemID].power;
	if(currentItemPower >= 0) {
        powerRequirement -= currentItemPower;
	}
	bool bhasPowerAfterwards = hasPower();

	if((currentGame->gameState != DEINITIALIZE) && (bhasPowerBefore != bhasPowerAfterwards)) {
        changeRadar(bhasPowerAfterwards);
	}

    // change spice capacity
	capacity -= currentGame->objectData.data[itemID].capacity;

    if(currentGame->gameState != LOADING) {
        // do not check selection lists if we are loading
        checkSelectionLists();
    }

	if (!isAlive())
		lose();

}


void PlayerClass::decrementHarvesters()
{
    numItem[Unit_Harvester]--;

    if (numItem[Unit_Harvester] <= 0) {
        numItem[Unit_Harvester] = 0;

        if(numItem[Structure_Refinery]) {
            Coord	closestPos;
            Coord	pos = Coord(0,0);
            double	closestDistance = INFINITY;
            StructureClass *closestRefinery = NULL;

            RobustList<StructureClass*>::const_iterator iter;
            for(iter = structureList.begin(); iter != structureList.end(); ++iter) {
                StructureClass* tempStructure = *iter;

                if((tempStructure->getItemID() == Structure_Refinery) && (tempStructure->getOwner() == this) && (tempStructure->getHealth() > 0.0)) {
                    pos = tempStructure->getLocation();

                    Coord closestPoint = tempStructure->getClosestPoint(pos);
                    double refineryDistance = blockDistance(pos, closestPoint);
                    if(!closestRefinery || (refineryDistance < closestDistance)) {
                            closestDistance = refineryDistance;
                            closestRefinery = tempStructure;
                            closestPos = pos;
                    }
                }
            }

            if(closestRefinery && (currentGame->gameState == BEGUN)) {
                freeHarvester(closestRefinery->getLocation().x, closestRefinery->getLocation().y);
            }
        }
    }
}

void PlayerClass::incrementUnits(int itemID)
{
    numUnits++;
    numItem[itemID]++;
}

void PlayerClass::decrementUnits(int itemID)
{
	numUnits--;

	if(itemID == Unit_Harvester) {
        decrementHarvesters();
	} else {
        numItem[itemID]--;
	}

	if (!isAlive())
		lose();

}


void PlayerClass::freeHarvester(int xPos, int yPos)
{
	if (currentGameMap->cellExists(xPos, yPos)
		&& currentGameMap->getCell(xPos, yPos)->hasAGroundObject()
		&& (currentGameMap->getCell(xPos, yPos)->getGroundObject()->getItemID() == Structure_Refinery))
	{
		RefineryClass* refinery = (RefineryClass*)currentGameMap->getCell(xPos, yPos)->getGroundObject();
		Coord closestPos = currentGameMap->findClosestEdgePoint(refinery->getLocation() + Coord(2,0), Coord(1,1));

		Carryall* carryall = (Carryall*)createUnit(Unit_Carryall);
		HarvesterClass* harvester = (HarvesterClass*)createUnit(Unit_Harvester);
		harvester->setAmountOfSpice(5.0);
		carryall->setOwned(false);
		carryall->giveCargo(harvester);
		carryall->deploy(closestPos);
		carryall->setDropOfferer(true);

		if (closestPos.x == 0)
			carryall->setAngle(RIGHT);
		else if (closestPos.x == currentGameMap->sizeX-1)
			carryall->setAngle(LEFT);
		else if (closestPos.y == 0)
			carryall->setAngle(DOWN);
		else if (closestPos.y == currentGameMap->sizeY-1)
			carryall->setAngle(UP);

		harvester->setTarget(refinery);
		harvester->setActive(false);
		carryall->setTarget(refinery);
	}
}


void PlayerClass::lose()
{
	char message[100];
	sprintf(message, "%s has been defeated.", getHouseNameByNumber( (HOUSETYPE) playerNumber).c_str());
	currentGame->AddToNewsTicker(message);

	if((getTeam() == thisPlayer->getTeam()) && ((currentGame->winFlags & WINLOSEFLAGS_HUMAN_HAS_BUILDINGS) != 0)) {
	    // thisPlayer is destroyed and this is a game finish condition
	    if((currentGame->loseFlags & WINLOSEFLAGS_HUMAN_HAS_BUILDINGS) != 0) {
	        // player has won
            currentGame->finished = true;
            currentGame->won = true;

            soundPlayer->playVoice(YouHaveFailedYourMission,house);
	    } else {
	        // player has lost
            currentGame->finished = true;
            currentGame->won = false;

            soundPlayer->playVoice(YourMissionIsComplete,house);
	    }

	} else if((currentGame->winFlags & WINLOSEFLAGS_AI_NO_BUILDINGS) != 0) {
		//if the only players left are on the thisPlayers team, thisPlayer has won
		bool finished = true;

		for(int i=0; i<NUM_HOUSES; i++) {
			if ((currentGame->player[i] != NULL) && currentGame->player[i]->isAlive()
				&& (currentGame->player[i]->getTeam() != 0) && (currentGame->player[i]->getTeam() != thisPlayer->getTeam()))
				finished = false;
		}

		if(finished) {
		    // all AI players are destroyed and this is a game finish condition
            if((currentGame->loseFlags & WINLOSEFLAGS_AI_NO_BUILDINGS) != 0) {
                // player has won
                currentGame->finished = true;
                currentGame->won = true;

                soundPlayer->playVoice(YouHaveFailedYourMission,house);
            } else {
                // player has lost
                currentGame->finished = true;
                currentGame->won = false;

                soundPlayer->playVoice(YourMissionIsComplete,house);
            }
		}
	}
}


void PlayerClass::noteDamageLocation(ObjectClass* pObject, const Coord& location)
{
	//if (this == thisPlayer)
	//	soundPlayer->playVOSound(Sound
}


void PlayerClass::returnCredits(double newCredits)
{
	if (newCredits > 0.0) {
		startingCredits += newCredits;
	}
}

void PlayerClass::setStartingCredits(double newCredits)
{
	if (newCredits >= 0.0)
		startingCredits = newCredits;
}

void PlayerClass::setStoredCredits(double newCredits)
{
	if (newCredits >= 0.0)
		credits = newCredits;
}

void PlayerClass::update()
{
	if (oldCredits != getAmountOfCredits()) {
		if ((this == thisPlayer) && (getAmountOfCredits() > 0))
			soundPlayer->playSound(CreditsTick);
		oldCredits = getAmountOfCredits();
	}

	if(credits > capacity) {
		credits--;// = capacity;
		if (this == thisPlayer)
			currentGame->AddToNewsTicker(pTextManager->getDuneText(DuneText_SpiceLost));
	}

	powerUsageTimer--;
	if(powerUsageTimer <= 0) {
	    powerUsageTimer = MILLI2CYCLES(15*1000);
        takeCredits(((double)getPowerRequirement()) / 32.0);
	}

	choam.update();
}

void PlayerClass::win()
{
	currentGame->finished = true;

	if (getTeam() == thisPlayer->getTeam())
	{
		currentGame->won = true;
		soundPlayer->playVoice(YourMissionIsComplete,house);
	}
	else
	{
		currentGame->won = false;
		soundPlayer->playVoice(YouHaveFailedYourMission,house);
	}
}

StructureClass* PlayerClass::placeStructure(Uint32 builderID, int itemID, int xPos, int yPos, bool bForcePlacing)
{
	if(!currentGameMap->cellExists(xPos,yPos)) {
		return NULL;
	}

	StructureClass* tempStructure = NULL;

	switch (itemID) {
		case (Structure_Slab1): {
			// Slabs are no normal buildings
			currentGameMap->getCell(xPos, yPos)->setType(Terrain_Slab);
			currentGameMap->getCell(xPos, yPos)->setOwner(getPlayerNumber());
			currentGameMap->viewMap(getTeam(), xPos, yPos, currentGame->objectData.data[Structure_Slab1].viewrange);
	//		currentGameMap->getCell(xPos, yPos)->clearTerrain();
			ObjectClass* pObject = currentGame->getObjectManager().getObject(builderID);
			if (pObject != NULL) {
				if (pObject->isAStructure() && ((StructureClass*) pObject)->isABuilder()) {
					((BuilderClass*) pObject)->unSetWaitingToPlace();
					if (this == thisPlayer) {
						currentGame->currentCursorMode = GameClass::CursorMode_Normal;
					}
				}
			} else if (this == thisPlayer) {
				currentGame->currentCursorMode = GameClass::CursorMode_Normal;
			}
		} break;

		case (Structure_Slab4): {
			// Slabs are no normal buildings
			int i,j;
			for(i = xPos; i < xPos + 2; i++) {
				for(j = yPos; j < yPos + 2; j++) {
					if (currentGameMap->cellExists(i, j)) {
						TerrainClass* cell = currentGameMap->getCell(i, j);

						if (!cell->hasAGroundObject() && cell->isRock() && !cell->isMountain()) {
							cell->setType(Terrain_Slab);
							cell->setOwner(getPlayerNumber());
							currentGameMap->viewMap(getTeam(), i, j, currentGame->objectData.data[Structure_Slab4].viewrange);
							//cell->clearTerrain();
						}
					}
				}
			}

			ObjectClass* pObject = currentGame->getObjectManager().getObject(builderID);
			if(pObject != NULL) {
				if (pObject->isAStructure() && ((StructureClass*) pObject)->isABuilder()) {
					((BuilderClass*) pObject)->unSetWaitingToPlace();

					if (this == thisPlayer) {
						currentGame->currentCursorMode = GameClass::CursorMode_Normal;
					}
				}
			} else if (this == thisPlayer) {
				currentGame->currentCursorMode = GameClass::CursorMode_Normal;
			}

		} break;

		default: {
			tempStructure = (StructureClass*) ObjectClass::createObject(itemID,this);
			if(tempStructure == NULL) {
				fprintf(stderr,"PlayerClass::placeStructure(): Cannot create Object with itemID %d\n",itemID);
				fflush(stderr);
				exit(EXIT_FAILURE);
			}

            if(bForcePlacing == false) {
                // check if there is already something on this cell
                for(int i=0;i<tempStructure->getStructureSizeX();i++) {
                    for(int j=0;j<tempStructure->getStructureSizeY();j++) {
                        if((currentGameMap->cellExists(xPos+i, yPos+j) == false) || (currentGameMap->getCell(xPos+i, yPos+j)->hasAGroundObject() == true)) {
                            delete tempStructure;
                            return NULL;
                        }
                    }
                }
            }

			for(int i=0;i<tempStructure->getStructureSizeX();i++) {
				for(int j=0;j<tempStructure->getStructureSizeY();j++) {
					if(currentGameMap->cellExists(xPos+i, yPos+j)) {
						currentGameMap->getCell(xPos+i, yPos+j)->clearTerrain();
					}
				}
			}

			tempStructure->setLocation(xPos, yPos);

			if ((builderID != NONE) && (itemID != Structure_Wall)) {
				tempStructure->setJustPlaced();
			}

			// at the beginning of the game every refinery gets one harvester for free (brought by a carryall)
			if(((currentGame->gameState == START) || (builderID != NONE)) && (itemID == Structure_Refinery)) {
				freeHarvester(xPos, yPos);
			}

			// if this structure was built by a construction yard this construction yard must be informed
			ObjectClass* pObject = currentGame->getObjectManager().getObject(builderID);
			if(pObject != NULL) {
				if(pObject->isAStructure() && ((StructureClass*) pObject)->isABuilder()) {
					((BuilderClass*) pObject)->unSetWaitingToPlace();
					if (this == thisPlayer) {
						currentGame->currentCursorMode = GameClass::CursorMode_Normal;
					}
				}
			} else if (this == thisPlayer) {
				currentGame->currentCursorMode = GameClass::CursorMode_Normal;
			}

			if(itemID == Structure_Radar && hasPower()) {
                changeRadar(true);
			}

			if(tempStructure->isABuilder()) {
                ((BuilderClass*) tempStructure)->checkSelectionList();
			}


		} break;
	}

	return tempStructure;
}



UnitClass* PlayerClass::createUnit(int itemID)
{
	UnitClass* newUnit = NULL;

	newUnit = (UnitClass*) ObjectClass::createObject(itemID,this);

	if(newUnit == NULL) {
		fprintf(stderr,"PlayerClass::createUnit(): Cannot create Object with itemID %d\n",itemID);
		fflush(stderr);
		exit(EXIT_FAILURE);
	}

	return newUnit;
}

UnitClass* PlayerClass::placeUnit(int itemID, int xPos, int yPos)
{
	UnitClass* newUnit = NULL;
	if(currentGameMap->cellExists(xPos, yPos) == true) {
	    TerrainClass* pTile = currentGameMap->getCell(xPos,yPos);

	    if(itemID == Unit_Fremen || itemID == Unit_Saboteur || itemID == Unit_Sardaukar || itemID == Unit_Soldier || itemID == Unit_Trooper) {
            if((pTile->hasANonInfantryGroundObject() == true) || (pTile->infantryNotFull() == false)) {
                // infantry units can not placed on non-infantry units or structures (or the cell is already full of infantry units)
                return NULL;
            }
	    } else {
	        if(pTile->hasAGroundObject() == true) {
                // non-infantry units can not placed on a cell where already some other unit or structure is placed on
                return NULL;
	        }
	    }

        newUnit = createUnit(itemID);
	}

	if (newUnit) {
		Coord pos = Coord(xPos, yPos);
		if (newUnit->canPass(xPos, yPos)) {
			newUnit->deploy(pos);
		} else {
			newUnit->setVisible(VIS_ALL, false);
			newUnit->destroy();
			newUnit = NULL;
		}
	}

	return newUnit;
}

double PlayerClass::takeCredits(double amount)
{
	double	taken = 0.0;

	if(getAmountOfCredits() >= 1) {
		if(credits >= amount) {
			taken = amount;
			credits -= amount;
		} else {
			taken = credits;
			credits = 0.0;

			if(startingCredits >= (amount - taken)) {
				startingCredits -= (amount - taken);
				taken = amount;
			} else {
				taken += startingCredits;
				startingCredits = 0.0;
			}
		}
	}

	return taken;	//the amount that was actually withdrawn
}

/**
    This method starts the animation for switching the radar on or off
    \param status   true if the radar is switched on, false if switched off
*/
void PlayerClass::changeRadar(bool status)
{
	if(this == thisPlayer) {
		soundPlayer->playSound(RadarNoise);

		if(status == true && thisPlayer->hasRadar()) {
			//turn on
			currentGame->radarView->switchRadarMode(true);
			soundPlayer->playVoice(RadarActivated,this->getHouse());
		} else {
			//turn off
			currentGame->radarView->switchRadarMode(false);
			soundPlayer->playVoice(RadarDeactivated,this->getHouse());
		}
	}
}

/**
    This method informs the player that one of his units has killed an enemy unit or structure
    \param itemID   the ID of the enemy unit or structure
*/
void PlayerClass::informHasKilled(Uint32 itemID)
{
    destroyedValue += std::max(currentGame->objectData.data[itemID].price/100, 1);
    if(isStructure(itemID)) {
        numDestroyedStructures++;
    } else {
        numDestroyedUnits++;
    }
}

	/**
        This method returns the center of the base of this player.
        \return the coordinate of the center in tile coordinates
	*/
Coord PlayerClass::getCenterOfMainBase() const {
    Coord center;
    int numStructures = 0;

    RobustList<StructureClass*>::const_iterator iter;
    for(iter = structureList.begin(); iter != structureList.end(); ++iter) {
        StructureClass* tempStructure = *iter;

        if(tempStructure->getOwner() == this) {
            center += tempStructure->getLocation();
            numStructures++;
        }
    }

    center /= numStructures;

    return center;
}

void PlayerClass::setProducedPower(int newPower)
{
	// calculate current power production
    bool bhasPowerBefore = hasPower();
	producedPower = newPower;
	bool bhasPowerAfterwards = hasPower();

    if(hasRadar() && (bhasPowerBefore != bhasPowerAfterwards)) {
        changeRadar(bhasPowerAfterwards);
	}
}
