/*
 *  This file is part of Dune Legacy.
 *
 *  Dune Legacy is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Dune Legacy is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Dune Legacy.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <misc/FileSystem.h>

#include <stdio.h>
#include <algorithm>

#ifdef _WIN32
#include <io.h>
#include <direct.h>
#else
#include <dirent.h>
#include <errno.h>
#include <string.h>
#endif

std::list<std::string> GetFileNames(std::string directory, std::string extension, bool IgnoreCase)
{

	std::list<std::string> Files;

	if(IgnoreCase == true) {
		std::transform(extension.begin(), extension.end(), extension.begin(), tolower);
	}

#ifdef _WIN32

	long hFile;

	_finddata_t fdata;

	std::string searchString = directory + "*";

	if ((hFile = (long)_findfirst(searchString.c_str(), &fdata)) != -1L) {
		while(_findnext(hFile, &fdata) == 0) {
			std::string Filename = fdata.name;
			unsigned int dotposition = Filename.find_last_of('.');

			if(dotposition == std::string::npos) {
				continue;
			}

			std::string ext = Filename.substr(dotposition+1);

			if(IgnoreCase == true) {
				std::transform(ext.begin(), ext.end(), ext.begin(), tolower);
			}

			if(ext == extension) {
				Files.push_back(Filename);
			}
		}

		_findclose(hFile);
	}

#else

	DIR * dir = opendir(directory.c_str());
	dirent *curEntry;

	if(dir == NULL) {
		return Files;
	}

	errno = 0;
	while((curEntry = readdir(dir)) != NULL) {
			std::string Filename = curEntry->d_name;
			size_t dotposition = Filename.find_last_of('.');

			if(dotposition == std::string::npos) {
				continue;
			}

			std::string ext = Filename.substr(dotposition+1);

			if(IgnoreCase == true) {
				std::transform(ext.begin(), ext.end(), ext.begin(), tolower);
			}

			if(ext == extension) {
				Files.push_back(Filename);
			}
	}

	if(errno != 0) {
		perror("readdir()");
	}

	closedir(dir);

#endif

	return Files;

}

bool GetCaseInsensitiveFilename(std::string& filepath) {

#ifdef _WIN32
    return ExistsFile(filepath);

#else


    std::string filename;
    std::string path;
    size_t separatorPos = filepath.rfind('/');
    if(separatorPos == std::string::npos) {
        // There is no '/' in the filepath => only filename in current working directory specified
        filename = filepath;
        path = ".";
    } else if(separatorPos == filepath.length()-1) {
        // filepath has an '/' at the end => no filename specified
        return false;
    } else {
        filename = filepath.substr(separatorPos+1, std::string::npos);
        path = filepath.substr(0,separatorPos+1); // path with tailing '/'
    }

    DIR* directory = opendir(path.c_str());
    if(directory == NULL) {
        return false;
    }

    while(true) {
        errno = 0;
        dirent* directory_entry = readdir(directory);
        if(directory_entry == NULL) {
            if (errno != 0) {
                closedir(directory);
                return false;
            } else {
                // EOF
                break;
            }
        }

        bool entry_OK = true;
        const char* pEntryName = directory_entry->d_name;
        const char* pFilename = filename.c_str();
        while(*pEntryName != '\0' && *pFilename != '\0') {
            if(tolower(*pEntryName) != tolower(*pFilename)) {
                entry_OK = false;
                break;
            }
            pEntryName++;
            pFilename++;
        }

        if(entry_OK == true) {
            if(path == ".") {
                filepath = directory_entry->d_name;
            } else {
                filepath = path + directory_entry->d_name;
            }
            closedir(directory);
            return true;
        }
    }
    closedir(directory);
    return false;

#endif

}


bool ExistsFile(const std::string path) {
	// try opening the file
	FILE* fp = fopen(path.c_str(),"r");

	if(fp == NULL) {
		return false;
	}

	fclose(fp);
	return true;
}
