/*
 *  This file is part of Dune Legacy.
 *
 *  Dune Legacy is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Dune Legacy is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Dune Legacy.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <structures/WallClass.h>

#include <MapClass.h>
#include <globals.h>

#include <FileClasses/GFXManager.h>
#include <PlayerClass.h>

WallClass::WallClass(PlayerClass* newOwner) : StructureClass(newOwner)
{
    WallClass::init();

    setHealth(getMaxHealth());

    bWallDestroyedUp = false;
    bWallDestroyedRight = false;
    bWallDestroyedDown = false;
    bWallDestroyedLeft = false;
}

WallClass::WallClass(Stream& stream) : StructureClass(stream)
{
    WallClass::init();

    bWallDestroyedUp = stream.readBool();
    bWallDestroyedRight = stream.readBool();
    bWallDestroyedDown = stream.readBool();
    bWallDestroyedLeft = stream.readBool();
}

void WallClass::init() {
    itemID = Structure_Wall;
	owner->incrementStructures(itemID);

	structureSize.x = 1;
	structureSize.y = 1;

	GraphicID = ObjPic_Wall;
	graphic = pGFXManager->getObjPic(GraphicID,getOwner()->getColour());
	imageW = graphic->w / 75;
	imageH = graphic->h;

	setWallTile(Wall_LeftRight);
}

WallClass::~WallClass()
{
}

void WallClass::save(Stream& stream) const
{
	StructureClass::save(stream);

    stream.writeBool(bWallDestroyedUp);
    stream.writeBool(bWallDestroyedRight);
    stream.writeBool(bWallDestroyedDown);
    stream.writeBool(bWallDestroyedLeft);
}

void WallClass::destroy()
{
	// fix wall to the north
	if(currentGameMap->cellExists(location.x, location.y-1) == true) {
        ObjectClass* obj = currentGameMap->getCell(location.x, location.y-1)->getGroundObject();
        if((obj != NULL) && (obj->getItemID() == Structure_Wall)) {
            ((WallClass*) obj)->bWallDestroyedDown = true;
            ((WallClass*) obj)->fixWall();
        }
	}

	// fix wall to the south
	if(currentGameMap->cellExists(location.x, location.y+1) == true) {
        ObjectClass* obj = currentGameMap->getCell(location.x, location.y+1)->getGroundObject();
        if((obj != NULL) && (obj->getItemID() == Structure_Wall)) {
            ((WallClass*) obj)->bWallDestroyedUp = true;
            ((WallClass*) obj)->fixWall();
        }
	}

	// fix wall to the west
	if(currentGameMap->cellExists(location.x-1, location.y) == true) {
        ObjectClass* obj = currentGameMap->getCell(location.x-1, location.y)->getGroundObject();
        if((obj != NULL) && (obj->getItemID() == Structure_Wall)) {
            ((WallClass*) obj)->bWallDestroyedRight = true;
            ((WallClass*) obj)->fixWall();
        }
	}

    // fix wall to the east
	if(currentGameMap->cellExists(location.x+1, location.y) == true) {
        ObjectClass* obj = currentGameMap->getCell(location.x+1, location.y)->getGroundObject();
        if((obj != NULL) && (obj->getItemID() == Structure_Wall)) {
            ((WallClass*) obj)->bWallDestroyedLeft = true;
            ((WallClass*) obj)->fixWall();
        }
	}


    StructureClass::destroy();
}

/**
    Sets the place for this wall and fixes the wall and their surounding walls.
    \param  xPos    the x position of this wall
    \param  yPos    the y position of this wall
*/
void WallClass::setLocation(int xPos, int yPos) {
    StructureClass::setLocation(xPos, yPos);

    // fix this wall
	fixWall();

	// fix wall to the north
	if(currentGameMap->cellExists(location.x, location.y-1) == true) {
        ObjectClass* obj = currentGameMap->getCell(location.x, location.y-1)->getGroundObject();
        if((obj != NULL) && (obj->getItemID() == Structure_Wall)) {
            ((WallClass*) obj)->bWallDestroyedDown = false;
            ((WallClass*) obj)->fixWall();
        }
	}

	// fix wall to the south
	if(currentGameMap->cellExists(location.x, location.y+1) == true) {
        ObjectClass* obj = currentGameMap->getCell(location.x, location.y+1)->getGroundObject();
        if((obj != NULL) && (obj->getItemID() == Structure_Wall)) {
            ((WallClass*) obj)->bWallDestroyedUp = false;
            ((WallClass*) obj)->fixWall();
        }
	}

	// fix wall to the west
	if(currentGameMap->cellExists(location.x-1, location.y) == true) {
        ObjectClass* obj = currentGameMap->getCell(location.x-1, location.y)->getGroundObject();
        if((obj != NULL) && (obj->getItemID() == Structure_Wall)) {
            ((WallClass*) obj)->bWallDestroyedRight = false;
            ((WallClass*) obj)->fixWall();
        }
	}

    // fix wall to the east
	if(currentGameMap->cellExists(location.x+1, location.y) == true) {
        ObjectClass* obj = currentGameMap->getCell(location.x+1, location.y)->getGroundObject();
        if((obj != NULL) && (obj->getItemID() == Structure_Wall)) {
            ((WallClass*) obj)->bWallDestroyedLeft = false;
            ((WallClass*) obj)->fixWall();
        }
	}
}

/**
    Fixes this wall. The choosen wall tile is based on the 4 surounding tiles and previously destroyed walls.
*/
void WallClass::fixWall() {

    int i = location.x;
    int j = location.y;

    int maketile = Wall_LeftRight;

    // Walls
    bool up = (!currentGameMap->cellExists(i, j-1) || (currentGameMap->cell[i][j-1].hasAGroundObject()
                && (currentGameMap->cell[i][j-1].getGroundObject()->getItemID() == Structure_Wall))) || bWallDestroyedUp;

    bool right = (!currentGameMap->cellExists(i+1, j) || (currentGameMap->cell[i+1][j].hasAGroundObject()
                && (currentGameMap->cell[i+1][j].getGroundObject()->getItemID() == Structure_Wall))) || bWallDestroyedRight;

    bool down = (!currentGameMap->cellExists(i, j+1) || (currentGameMap->cell[i][j+1].hasAGroundObject()
                && (currentGameMap->cell[i][j+1].getGroundObject()->getItemID() == Structure_Wall))) || bWallDestroyedDown;

    bool left = (!currentGameMap->cellExists(i-1, j) || (currentGameMap->cell[i-1][j].hasAGroundObject()
                && (currentGameMap->cell[i-1][j].getGroundObject()->getItemID() == Structure_Wall))) || bWallDestroyedLeft;

    // calculate destroyed tile index
    int destroyedTileIndex = 0;
    if(left == true)
        destroyedTileIndex = (destroyedTileIndex << 1) | bWallDestroyedLeft;

    if(down == true)
        destroyedTileIndex = (destroyedTileIndex << 1) | bWallDestroyedDown;

    if(right == true)
        destroyedTileIndex = (destroyedTileIndex << 1) | bWallDestroyedRight;

    if(up == true)
        destroyedTileIndex = (destroyedTileIndex << 1) | bWallDestroyedUp;

    // Now perform the test
    if ((left == true) && (right == true) && (up == true) && (down == true)) {
        maketile = (destroyedTileIndex == 0) ? Wall_Full : 59 + destroyedTileIndex; //solid wall
    } else if ((left == false) && (right == true) && (up == true) && (down == true)) {
        maketile = (destroyedTileIndex == 0) ? Wall_UpDownRight : 31 + destroyedTileIndex; //missing left edge
    } else if ((left == true) && (right == false)&& (up == true) && (down == true)) {
        maketile = (destroyedTileIndex == 0) ? Wall_UpDownLeft : 45 + destroyedTileIndex; //missing right edge
    } else if ((left == true) && (right == true) && (up == false) && (down == true)) {
        maketile = (destroyedTileIndex == 0) ? Wall_DownLeftRight : 38 + destroyedTileIndex; //missing top edge
    } else if ((left == true) && (right == true) && (up == true) && (down == false)) {
        maketile = (destroyedTileIndex == 0) ? Wall_UpLeftRight : 52 + destroyedTileIndex; //missing bottom edge
    } else if ((left == false) && (right == true) && (up == false) && (down == true)) {
        maketile = (destroyedTileIndex == 0) ? Wall_DownRight : 22 + destroyedTileIndex; //missing top left edge
    } else if ((left == true) && (right == false) && (up == true) && (down == false)) {
        maketile = (destroyedTileIndex == 0) ? Wall_UpLeft : 28 + destroyedTileIndex; //missing bottom right edge
    } else if ((left == true) && (right == false) && (up == false) && (down == true)) {
        maketile = (destroyedTileIndex == 0) ? Wall_DownLeft : 25 + destroyedTileIndex; //missing top right edge
    } else if ((left == false) && (right == true) && (up == true) && (down == false)) {
        maketile = (destroyedTileIndex == 0) ? Wall_UpRight : 19 + destroyedTileIndex; //missing bottom left edge
    } else if ((left == true) && (right == false) && (up == false) && (down == false)) {
        maketile = (destroyedTileIndex == 0) ? Wall_LeftRight : 14 + destroyedTileIndex; //missing above, right and below
    } else if ((left == false) && (right == true) && (up == false) && (down == false)) {
        maketile = (destroyedTileIndex == 0) ? Wall_LeftRight : 13 + destroyedTileIndex; //missing above, left and below
    } else if ((left == false) && (right == false) && (up == true) && (down == false)) {
        maketile = (destroyedTileIndex == 0) ? Wall_UpDown : 16 + destroyedTileIndex; //only up
    } else if ((left == false) && (right == false) && (up == false) && (down == true)) {
        maketile = (destroyedTileIndex == 0) ? Wall_UpDown : 17 + destroyedTileIndex; //only down
    } else if ((left == true) && (right == true) && (up == false) && (down == false)) {
        maketile = (destroyedTileIndex == 0) ? Wall_LeftRight : 13 + destroyedTileIndex; //missing above and below
    } else if ((left == false) && (right == false) && (up == true) && (down == true)) {
        maketile = (destroyedTileIndex == 0) ? Wall_UpDown : 16 + destroyedTileIndex; //missing left and right
    } else if ((left == false) && (right == false) && (up == false) && (down == false)) {
        maketile = (destroyedTileIndex == 0) ? Wall_Standalone : 12 + destroyedTileIndex; //missing left and right
    }

    setWallTile(maketile);
}
